--                                                                    --
--  package Strings_Edit            Copyright (c)  Dmitry A. Kazakov  --
--  Interface                                      Luebeck            --
--                                                 Spring, 2000       --
--                                                                    --
--  This  library  is  free software; you can redistribute it and/or  --
--  modify it under the terms of  the  GNU  Library  General  Public  --
--  License  as  published  by  the Free Software Foundation; either  --
--  version 2 of the License, or (at your option) any later version.  --
--                                                                    --
--  This  library is distributed in the hope that it will be useful,  --
--  but WITHOUT ANY WARRANTY; without even the implied  warranty  of  --
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  --
--  Library General Public License for more details.                  --
--                                                                    --
--  You  should  have  received  a  copy  of the GNU Library General  --
--  Public  License  along  with  this library; if not, write to the  --
--  Free Software Foundation, Inc.,  675  Mass  Ave,  Cambridge,  MA  --
--  02139, USA.                                                       --
--____________________________________________________________________--
--
-- The following I/O items are supported by the package:
--
-- (o) Integer numbers (generic, package Integer_Edit)
-- (o) Floating-point numbers (generic, package Float_Edit)
-- (o) Roman numbers (the type Roman)
-- (o) Strings
--
-- The major differences to  the  standard  Image/Value  attributes  and
-- Text_IO procedures.
--
-- 1. For  numeric  types,  the  base  is  neither written nor read. For
--    instance, output of 23 as hexadecimal gives 17, not 16#17#. 
--
-- 2. Get procedures do not skip blank characters around  input  tokens,
--    except the cases when the blank charactes is required by syntax. 
--
-- 3. Get  procedures  use  the current string position pointer, so that
--    they can be consequently called advancing the pointer as the tokes
--    are recognized. 
--
-- 4. Numeric  get procedures allow to specify the expected value range.
--    When  the  actual  value  is  out  of  the range then depending on
--    procedure  parameters, either Constrain_Error is propagated or the
--    value is forced to the nearest range boundary. 
--
-- 5. Put procedures also use the current string position pointer, which
--    allow to call them consequently.
--
-- 6. The format used for floating-number output is based on the  number
--    precision, instead of rather typographic approach of Text_IO.  The
--    precision can be specified either as the number of valid digits of
--    the  current  base  (i.e. relative) or as the position of the last
--    valid digit (i.e. absolute). For instance, 12.345678 with relative
--    precision 3 gives 12.3.  With  absolute  precision  -3,  it  gives
--    12.346. 
--
-- G E T   P R O C E D U R E S   
--
-- Get procedures are used to scan strings. The first two parameters are
-- always Source and Pointer. The Source is the string to be scaned. The
-- Pointer indicates the current position. After  successful  completion
-- it is advanced to the first string position following the  recognized
-- item.  The  pointer  shall  be   in   the   range   Source'First   ..
-- Source'Last+1. The Layout_Error exception  is  propagated  when  this
-- check fails. The third  parameter  usually  accepts  the  value.  The
-- following example shows how to use get procedures: 
--
--       package Edit_Float is new Float_Edit (Float);
--       use Edit_Float;
--
--       Line        : String (1..512);   -- A line
--       Pointer     : Integer;
--       Value       : Float;
--       TabAndSpace : Ada.Strings.Maps.Character_Set :=
--                        To_Set (" " & Ada.Characters.Latin_1.HT);
--    begin
--       ...
--       Pointer := Line'First;
--       Get (Line, Pointer, TabAndSpace);  -- Skip tabs and spaces
--       Get (Line, Pointer, Value);        -- Get number
--       Get (Line, Pointer, TabAndSpace);  -- Skip tabs and spaces
--       ... 
--
-- Numeric Get procedures have additional controlling the range  of  the
-- input value. Parameters First  and  Last  define  the  range  of  the
-- expected value. The exception Constraint_Error is propagated when the
-- value is not in the range. The exception can be suppressed using  the
-- parameters ToFirst and ToLast, which cause  the  input  value  to  be
-- substituted by the corresponding margin when the parameter is True. 
--
-- V A L U E   F U N C T I O N S
--
-- Each get procedure returning some value  has  a  corresponding  Value
-- function. The Value function has the same parameter profile with  the
-- exception  that  the  parameter  Pointer  is  absent and the value is
-- returned via result. Unlike Get procedure  Value  function  tolerates
-- spaces  and  tabs around the converted value. The whole string should
-- be matched, otherwise, Data_Error is propagated. 
--
-- P U T   P R O C E D U R E S
--
-- Put procedures place something into the  output  string  Destination.
-- The string  is  written  starting  from  Destination  (Pointer).  The
-- parameter Field defines the output size. When it has the  value  zero
-- then the output size is defined by the  output  item.  Otherwise  the
-- output  is  justified  within  the  field  and  the parameter Justify
-- specifies output alignment and  the  Fill  parameter  gives  the  pad
-- character. When Field is greater than Destination'Last - Pointer + 1,
-- the  latter  is  used instead. After successful completion Pointer is
-- advanced   to   the  first  character  following  the  output  or  to
-- Destination'Last + 1. 
--
-- I M A G E   F U N C T I O N S
--
-- Image  functions convert a value into string. Unlike standard S'Image
-- they do not place an extra space character. 
--
with IO_Exceptions;
with Ada.Strings;
with Ada.Strings.Maps;
with Ada.Characters.Latin_1;

package Strings_Edit is
   MaxSmall     : constant := 250;	-- Bigger than any possible

   subtype Alignment is Ada.Strings.Alignment;

   Center : Alignment renames Ada.Strings.Center;
   Left   : Alignment renames Ada.Strings.Left;
   Right  : Alignment renames Ada.Strings.Right;

   type Roman is range 1..3999;

   subtype NumberBase is Integer range 2..16;

   SpaceAndTab : constant Ada.Strings.Maps.Character_Set :=
                    Ada.Strings.Maps.To_Set
                    (  ' '
                    &  Ada.Characters.Latin_1.HT
                    );
--
-- S T R I N G S
--
   -- Get -- Skip blank characters
   --
   --    Source  - The string to be processed
   --    Pointer - The current position in the string
   --    Blank   - The blank character
   --    
   -- This procedure skips the  character  Blank  starting  from  Source
   -- (Pointer).  The  Pointer  is  advanced  to  the  first   non-Blank
   -- character or to Source'Last + 1. 
   --
   -- Exceptions :
   --
   --    Layout_Error - Pointer is not in Source'First..Source'Last + 1
   --
   procedure Get
             (  Source  : in String;
                Pointer : in out Integer;
                Blank   : in Character := ' '
             );
   --
   -- Get -- Skip blank characters
   --
   --    Source  - The string to be processed
   --    Pointer - The current position in the string
   --    Blanks  - The set characters to be considered as blank ones
   --    
   -- This procedure skips all the characters of the set Blanks starting
   -- from  Source  (Pointer).  The  Pointer  is  advanced  to the first
   -- non-blank character or to Source'Last + 1. 
   --
   -- Exceptions :
   --
   --    Layout_Error - Pointer is not in Source'First..Source'Last + 1
   --
   procedure Get
             (  Source  : in String;
                Pointer : in out Integer;
                Blanks  : in Ada.Strings.Maps.Character_Set
             );
   --
   -- Put -- Put a character into a string
   --
   --    Destination - The string that accepts the output
   --    Pointer     - The current position in the string
   --    Value       - The character to be put
   --    Field       - The output field
   --    Justify     - Alignment within the field
   --    Fill        - The fill character
   --
   -- This  procedure  places  the specified character (Value parameter)
   -- into the output string Destination. The string is written starting
   -- from the Destination (Pointer).
   --
   -- Exceptions:
   --
   --    Layout_Error -- Pointer is not in Destination'Range or there is
   --                    no room for the output. 
   --
   procedure Put
             (  Destination : in out String;
                Pointer     : in out Integer;
                Value       : in Character;
                Field       : in Natural := 0;
                Justify     : in Alignment := Left;
                Fill        : in Character := ' '
             );
   --
   -- Put -- Put a string into another string
   --
   --    Destination - The string that accepts the output
   --    Pointer     - The current position in the string
   --    Value       - The string to be put
   --    Field       - The output field
   --    Justify     - Alignment within the field
   --    Fill        - The fill character
   --
   -- This  procedure places the specified by the Value parameter string
   -- into the output string Destination. The string is written starting
   -- from the Destination (Pointer).
   --
   -- Exceptions:
   --
   --    Layout_Error -- Pointer is not in Destination'Range or there is
   --                    no room for the output. 
   --
   procedure Put
             (  Destination : in out String;
                Pointer     : in out Integer;
                Value       : in String;
                Field       : in Natural := 0;
                Justify     : in Alignment := Left;
                Fill        : in Character := ' '
             );
--
-- R O M A N   N U M B E R S
--
   -- Get -- Get a roman number from the string
   --
   --    Source  - The string to be processed
   --    Pointer - The current position in the string
   --    Value   - The result
   --    First   - The lowest allowed value
   --    Last    - The highest allowed value
   --    ToFirst - Force value to First instead of exception
   --    ToLast  - Force value to Last instead of exception
   --
   -- This  procedure  gets  a  roman number from the string Source. The
   -- process starts from the Source (Pointer) position. 
   --
   -- Exceptions:
   --
   --    Constraint_Error - The number is not in First..Last
   --    Data_Error       - Syntax error in the number
   --    End_Error        - There is no any number
   --    Layout_Error     - Pointer not in Source'First .. Source'Last +
   --                       1 
   --
   procedure Get
             (  Source  : in String;
                Pointer : in out Integer;
                Value   : out Roman;
                First   : in Roman := Roman'First;
                Last    : in Roman := Roman'Last;
                ToFirst : in Boolean := False;
                ToLast  : in Boolean := False
             );
   --
   -- Value -- String to roman number conversion
   --
   --    Source  - The string to be processed
   --    First   - The lowest allowed value
   --    Last    - The highest allowed value
   --    ToFirst - Force value to First instead of exception
   --    ToLast  - Force value to Last instead of exception
   --
   -- This function gets the roman number from the  string  Source.  The
   -- number can be surrounded by spaces  and  tabs.  The  whole  Source
   -- string should be matched. Otherwise the  exception  Data_Error  is
   -- propagated. 
   --
   -- Returns :
   --
   --    The value
   --
   -- Exceptions:
   --
   --    Constraint_Error - The number is not in First..Last
   --    Data_Error       - Syntax error in the number
   --    End_Error        - There is no any number
   --
   function Value
            (  Source  : in String;
               First   : in Roman := Roman'First;
               Last    : in Roman := Roman'Last;
               ToFirst : in Boolean := False;
               ToLast  : in Boolean := False
            )  return Roman;
   --
   -- Put -- Put a roman number into another string
   --
   --    Destination - The string that accepts the output
   --    Pointer     - The current position in the string
   --    Value       - The number to be put
   --    LowerCase   - Use lower case letters for the output
   --    Field       - The output field
   --    Justify     - Alignment within the field
   --    Fill        - The fill character
   --
   -- This  procedure places the number specified by the parameter Value
   -- into the output string Destination. The string is written starting
   -- from the Destination (Pointer). The parameter LowerCase determines
   -- whether upper or lower case letter should be used. 
   --
   -- Exceptions:
   --
   --    Layout_Error -- Pointer is not in Destination'Range or there is
   --                    no room for the output. 
   --
   procedure Put
             (  Destination : in out String;
                Pointer     : in out Integer;
                Value       : in Roman;
                LowerCase   : in Boolean := False;
                Field       : in Natural := 0;
                Justify     : in Alignment := Left;
                Fill        : in Character := ' '
             );
   --
   -- Image -- Roman number to string conversion
   --
   --    Value      - The value to be converted
   --    LowerCase  - Use lower case letters for the output
   --
   -- This function converts Value to string.  The  parameter  LowerCase
   -- indicates whether upper or lower case letters shall be used.
   --
   -- Returns :
   --
   --    The string
   --
   function Image
            (  Value     : in Roman;
               LowerCase : in Boolean := False
            )  return String;
--
-- I N T E G E R   N U M B E R S
--
   -- Integer_Edit -- Generic package for integer numbers
   --
   generic
      type Number is range <>;
   package Integer_Edit is
      --
      -- Get -- Get an integer number from the string
      --
      --    Source  - The string to be processed
      --    Pointer - The current position in the string
      --    Value   - The result
      --    Base    - The base of the expected number
      --    First   - The lowest allowed value
      --    Last    - The highest allowed value
      --    ToFirst - Force value to First instead of exception
      --    ToLast  - Force value to Last instead of exception
      --
      -- This procedure gets an integer number from the  string  Source.
      -- The process starts from the Source (Pointer) position. 
      --
      -- Exceptions:
      --
      --    Constraint_Error - The number is not in First..Last
      --    Data_Error       - Syntax error in the number
      --    End_Error        - There is no any number
      --    Layout_Error     - Pointer not in Source'First .. Source'Last
      --                       + 1
      --
      procedure Get
                (  Source  : in String;
                   Pointer : in out Integer;
                   Value   : out Number'Base;
                   Base    : in NumberBase := 10;
                   First   : in Number'Base := Number'First;
                   Last    : in Number'Base := Number'Last;
                   ToFirst : in Boolean := False;
                   ToLast  : in Boolean := False
                );
      --
      -- Value -- String to an integer conversion
      --
      --    Source  - The string to be processed
      --    Base    - The base of the expected number
      --    First   - The lowest allowed value
      --    Last    - The highest allowed value
      --    ToFirst - Force value to First instead of exception
      --    ToLast  - Force value to Last instead of exception
      --
      -- This  function  gets  an integer number from the string Source.
      -- The number can be surrounded by  spaces  and  tabs.  The  whole
      -- Source  string  should  be  matched.  Otherwise  the  exception
      -- Data_Error is propagated. 
      --
      -- Returns :
      --
      --    The value
      --
      -- Exceptions:
      --
      --    Constraint_Error - The number is not in First..Last
      --    Data_Error       - Syntax error in the number
      --    End_Error        - There is no any number
      --
      function Value
               (  Source  : in String;
                  Base    : in NumberBase := 10;
                  First   : in Number'Base := Number'First;
                  Last    : in Number'Base := Number'Last;
                  ToFirst : in Boolean := False;
                  ToLast  : in Boolean := False
               )  return Number'Base;
      --
      -- Put -- Put an integer into a string
      --
      --    Destination - The string that accepts the output
      --    Pointer     - The current position in the string
      --    Value       - The value to be put
      --    Base        - The base used for the output
      --    PutPlus     - The plus should placed for positive numbers
      --    Field       - The output field
      --    Justify     - Alignment within the field
      --    Fill        - The fill character
      --
      -- This  procedure  places  the  number  specified  by  the  Value
      -- parameter  into  the  output  string Destination. The string is
      -- written  starting from the Destination (Pointer). The parameter
      -- Base indicates the number base used for the  output.  The  base
      -- itself does not appear in the  output.  The  parameter  PutPlus
      -- indicates whether the plus sign should be placed if the  number
      -- is positive. 
      --
      -- Exceptions:
      --
      --    Layout_Error -- Pointer is not in Destination'Range or there
      --                    is no room for the output. 
      --
      -- Example :
      --
      --    Text    : String (1..20) := (others =>'#');
      --    Pointer : Positive := Text'First;
      --
      --    Put (Text, Pointer, 5, 2, True, 10, Center, '@');
      --
      --    Now the Text is "@@@+101@@@##########" Pointer = 11
      --
      procedure Put
                (  Destination : in out String;
                   Pointer     : in out Integer;
                   Value       : in Number'Base;
                   Base        : in NumberBase := 10;
                   PutPlus     : in Boolean := False;
                   Field       : in Natural := 0;
                   Justify     : in Alignment := Left;
                   Fill        : in Character := ' '
                );
      --
      -- Image -- Integer to string conversion
      --
      --    Value       - The value to be converted
      --    Base        - The base used for the output
      --    PutPlus     - The plus should placed for positive numbers
      --
      -- This function converts Value  to  string.  The  parameter  Base
      -- indicates the number base used for the output. The base  itself
      -- does  not  appear  in the output. The PutPlus indicates whether
      -- the plus sign should be placed if the number is positive. 
      --
      -- Returns :
      --
      --    The result string
      --
      function Image
               (  Value   : in Number'Base;
                  Base    : in NumberBase := 10;
                  PutPlus : in Boolean := False
               )  return String;
   end Integer_Edit;
--
-- F L O A T I N G - P O I N T   N U M B E R S
--
   -- Float_Edit -- Generic package for floating-point numbers
   --
   generic
      type Number is digits <>;
   package Float_Edit is
      --
      -- Get -- Get a floating-point number from the string
      --
      --    Source  - The string to be processed
      --    Pointer - The current position in the string
      --    Value   - The result
      --    Base    - The base of the expected number
      --    First   - The lowest allowed value
      --    Last    - The highest allowed value
      --    ToFirst - Force value to First instead of exception
      --    ToLast  - Force value to Last instead of exception
      --
      -- This  procedure  gets  a  number  from  the  string Source. The
      -- process  starts  from the Source (Pointer) position. The number
      -- in  the  string  may be in either floating-point or fixed-point
      -- format.  The  point  may  be absent. The mantissa can have base
      -- 2..16  (defined  by  the parameter Base). The exponent part (if
      -- appears) is introduced by 'e' or 'E'. It is always  decimal  of
      -- Base  radix.  Space characters are allowed between mantissa and
      -- exponent  part  and  in  the  exponent part around the exponent
      -- sign. If Base has the value 15 or 16 the exponent part shall be
      -- separated by at least one space character from mantissa. 
      -- 
      -- Exceptions:
      --
      --    Constraint_Error - The number is not in First..Last
      --    Data_Error       - Syntax error in the number
      --    End_Error        - There is no any number
      --    Layout_Error     - Pointer not in  Source'First..Source'Last
      --                       + 1
      --
      procedure Get
                (  Source   : in String;
                   Pointer  : in out Integer;
                   Value    : out Number'Base;
                   Base     : in NumberBase := 10;
                   First    : in Number'Base := Number'Base'First;
                   Last     : in Number'Base := Number'Base'Last;
                   ToFirst  : in Boolean := False;
                   ToLast   : in Boolean := False
                );
      --
      -- Value -- String to floating-point conversion
      --
      --    Source  - The string to be processed
      --    Base    - The base of the expected number
      --    First   - The lowest allowed value
      --    Last    - The highest allowed value
      --    ToFirst - Force value to First instead of exception
      --    ToLast  - Force value to Last instead of exception
      --
      -- This function gets the floating-point number  from  the  string
      -- Source. The number can be surrounded by spaces  and  tabs.  The
      -- whole Source string should be matched. Otherwise the  exception
      -- Data_Error is propagated. 
      --
      -- Returns :
      --
      --    The value
      --
      -- Exceptions:
      --
      --    Constraint_Error - The number is not in First..Last
      --    Data_Error       - Syntax error in the number
      --    End_Error        - There is no any number
      --
      function Value
               (  Source  : in String;
                  Base    : in NumberBase := 10;
                  First   : in Number'Base := Number'First;
                  Last    : in Number'Base := Number'Last;
                  ToFirst : in Boolean := False;
                  ToLast  : in Boolean := False
               )  return Number'Base;
      --
      -- Put -- Put a floating-point number into a string
      --
      --    Destination - The string that accepts the output
      --    Pointer     - The current position in the string
      --    Value       - The value to be put
      --    Base        - The base used for the output
      --    PutPlus     - The plus should placed for positive numbers
      --    RelSmall    - Relative precision of the output
      --    AbsSmall    - Absolute one
      --    Field       - The output field
      --    Justify     - Alignment within the field
      --    Fill        - The fill character
      --
      -- This  procedure  places  the  number specified by the parameter
      -- Value into the output string Destination. The string is written
      -- starting   from   Destination  (Pointer).  The  parameter  Base
      -- indicates the number base used for the output. Base itself does
      -- not appear in the output. The exponent part (if used) is always
      -- decimal. PutPlus indicates whether  the  plus  sign  should  be
      -- placed if the number is positive. There are two ways to specify
      -- the output precision. 
      --
      -- (o)  The parameter RelSmall determines the number of the  right
      --      Base places of the mantissa. For instance, with RelSmall =
      --      3 and Base = 10, the number 1.234567e+01 is represented as
      --      12.3. 
      --               
      -- (o)  The parameter AbsSmall determines  the  rightmost  correct
      --      digit of the mantissa. For example,  with  AbsSmall  =  0,
      --      Base  =  10,  the number 1.234567e+01 is represented as 12
      --      (i.e. 12.34567 rounded to 10**0). 
      --
      -- From  two  parameters  RelSmall  and  AbsSmall,  the  procedure
      -- chooses one, that specifies  the  minimal  number  of  mantissa
      -- digits,  but  no  more  than  the machine representation of the
      -- number allows. If the  point  would  appear  in  the  rightmost
      -- position  it  is omited. The pure zero is always represented as
      -- 0. If the desired number of  digits  may  be  provided  in  the
      -- fixed-point  format  then  the  exponent  part is not used. For
      -- example,  1.234567e-04  gives  0.0001234567  because fixed- and
      -- floating-point formats have the same length.  But  1.234567e-05
      -- will be shown in the floating-point format. For bases 15 and 16
      -- the  exponent  part is separated from the mantissa by space (to
      -- avoid ambiguity: F.Ee+2 is F.EE +  2  or  F.E  *  16**2?).  The
      -- parameter Field defines the output size. If it  has  the  value
      -- zero, then the output field is equal to the output length.
      --
      -- When the parameter Field is not  zero  then  Justify  specifies
      -- alignment  and  Fill  is  the  character used for filling. When
      -- Field is greater than  Destination'Last  -  Pointer  +  1,  the
      -- latter is used instead. After successful completion Pointer  is
      -- advanced  to  the  first  character  following the output or to
      -- Destination'Last + 1. 
      --
      -- Exceptions:
      --
      --    Layout_Error -- Pointer is not in Destination'Range or there
      --                    is no room for the output. 
      --
      procedure Put
                (  Destination : in out String;
                   Pointer     : in out Integer;
                   Value       : in Number'Base;
                   Base        : in NumberBase := 10;
                   PutPlus     : in Boolean := False;
                   RelSmall    : in Positive := MaxSmall;
                   AbsSmall    : in Integer := -MaxSmall;
                   Field       : in Natural := 0;
                   Justify     : in Alignment := Left;
                   Fill        : in Character := ' '
                );
      --
      -- Image -- Floating-point to string conversion
      --
      --    Value       - The value to be converted
      --    Base        - The base used for the output
      --    PutPlus     - The plus should placed for positive numbers
      --    RelSmall    - Relative precision of the output
      --    AbsSmall    - Absolute one
      --
      -- This procedure converts the parameter Value. The Base parameter
      -- indicates the number base used for the output. The Base  itself
      -- does not appear in the output. The exponent part (if  used)  is
      -- always  decimal.  The  PutPlus  indicates whether the plus sign
      -- should be placed if  the  number  is  positive.  For  precision
      -- parameters see Put. 
      --
      -- Returns :
      --
      --	 The result string
      --
      function Image
               (  Value    : in Number'Base;
                  Base     : in NumberBase := 10;
                  PutPlus  : in Boolean := False;
                  RelSmall : in Positive := MaxSmall;
                  AbsSmall : in Integer := -MaxSmall
               )  return String;
   end Float_Edit;
--
-- T R I M   F U N C T I O N S
--
   -- Trim -- Delete blank characters form string ends
   --
   --    Source - The string to be processed
   --    Blank  - The blank character
   --
   -- This function removes the Blank character from both  ends  of  the
   -- string and returns the result. 
   --
   -- Returns :
   --
   --    The result string
   --
   function Trim
            (  Source : in String;
               Blank  : in Character := ' '
            )  return String;
   --
   -- Trim -- Delete blank characters form string ends
   --
   --    Source - The string to be processed
   --    Blanks - The set of blank characters
   --
   -- This  function  removes any characters of the Blanks set from both
   -- ends of the string and returns the result. 
   --
   -- Returns :
   --
   --    The result string
   --
   function Trim
            (  Source : in String;
               Blanks : in Ada.Strings.Maps.Character_Set
            )  return String;
end Strings_Edit;
